#!/usr/bin/env python

"""
A simple (and sane) text indexing library.

Copyright (C) 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from os.path import commonprefix # to find common string prefixes
from bisect import bisect_right  # to find terms in the dictionary index

# Foundation classes.

class File:

    "A basic file abstraction."

    def __init__(self, f):
        self.f = f
        self.reset()

    def reset(self):
        pass

    def close(self):
        self.f.close()

class FileWriter(File):

    "Writing basic data types to files."

    def write_number(self, number):

        "Write 'number' to the file using a variable length encoding."

        # Negative numbers are not supported.

        if number < 0:
            raise ValueError, "Number %r is negative." % number

        # Special case: one byte containing zero.

        elif number == 0:
            self.f.write(chr(1) + chr(0))
            return

        # Write the number from least to most significant digits.

        nbytes = 0
        bytes = []

        while number != 0:
            lsd = number & 255
            bytes.append(chr(lsd))
            number = number >> 8
            nbytes += 1

        # Too large numbers are not supported.

        if nbytes > 255:
            raise ValueError, "Number %r is too large." % number

        bytes.insert(0, chr(nbytes))
        record = "".join(bytes)
        self.f.write(record)

    def write_unsigned_byte(self, number):

        "Write 'number' to the file using a single byte."

        if not (0 <= number <= 255):
            raise ValueError, "Number %r is out of range." % number

        self.f.write(chr(number))

    def write_string(self, s):

        "Write 's' to the file, recording its length."

        length = len(s)

        if not (0 <= length <= 255):
            raise ValueError, "String %r is too long." % s

        self.write_unsigned_byte(length)
        self.f.write(s)

class FileReader(File):

    "Reading basic data types from files."

    def read_number(self):

        "Read a number from the file."

        nbytes = self.read_unsigned_byte()

        # Read each byte, adding it to the number.

        bytes = self.f.read(nbytes)

        i = 0
        shift = 0
        number = 0

        while i < nbytes:
            csd = ord(bytes[i])
            number += (csd << shift)
            shift += 8
            i += 1

        return number

    def read_unsigned_byte(self):

        "Read a number from the file, consuming a single byte."

        s = self.f.read(1)
        if not s:
            raise EOFError

        return ord(s)

    def read_string(self):

        "Read a string from the file."

        length = self.read_unsigned_byte()
        return self.f.read(length)

# Specific classes.

class PositionWriter(FileWriter):

    "Writing position information to files."

    def reset(self):
        self.last_docnum = 0

    def write_positions(self, docnum, positions):

        "Write for the document 'docnum' the given 'positions'."

        if docnum < self.last_docnum:
            raise ValueError, "Document number %r is less than previous number %r." % (docnum, self.last_docnum)

        # Write the document number delta.

        self.write_number(docnum - self.last_docnum)

        # Write the number of positions.

        self.write_number(len(positions))

        # Write the position deltas.

        last = 0
        for position in positions:
            pos = position - last
            self.write_number(pos)
            last = position

        self.last_docnum = docnum

    def write_all_positions(self, doc_positions):

        """
        Write all 'doc_positions' - a collection of tuples of the form (document
        number, position list) - to the file, returning the offset at which they
        were stored.
        """

        # Reset the writer and record the current file offset.

        self.reset()
        offset = self.f.tell()

        # Write the number of documents.

        self.write_number(len(doc_positions))

        # Write the positions.

        for docnum, positions in doc_positions:
            self.write_positions(docnum, positions)

        return offset

class PositionReader(FileReader):

    "Reading position information from files."

    def reset(self):
        self.last_docnum = 0

    def read_positions(self):

        "Read positions, returning a document number and a list of positions."

        # Read the document number delta and add it to the last number.

        self.last_docnum += self.read_number()

        # Read the number of positions.

        npositions = self.read_number()

        # Read the position deltas, adding each previous position to get the
        # appropriate collection of absolute positions.

        i = 0
        last = 0
        positions = []

        while i < npositions:
            last += self.read_number()
            positions.append(last)
            i += 1

        return self.last_docnum, positions

    def read_all_positions(self, offset):

        """
        Read all positions from 'offset', seeking to that position in the file
        before reading.
        """

        self.reset()
        self.f.seek(offset)

        # Read the number of documents.

        ndocuments = self.read_number()

        # Read all records.

        i = 0
        doc_positions = []

        while i < ndocuments:
            doc_positions.append(self.read_positions())
            i += 1

        return doc_positions

class TermWriter(FileWriter):

    "Writing term information to files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def write_term(self, term, offset):

        """
        Write the given 'term' and its position file 'offset' to the term
        information file. Return the offset after the term information was
        written to the file.
        """

        # Too long terms are not currently supported.

        if len(term) > 255:
            raise ValueError, "Term %r is too long." % term

        # Write the prefix length and term suffix.

        common = len(commonprefix([self.last_term, term]))
        suffix = term[common:]

        self.write_unsigned_byte(common)
        self.write_string(suffix)

        # Write the offset delta.

        self.write_number(offset - self.last_offset)

        self.last_term = term
        self.last_offset = offset

        return self.f.tell()

class TermReader(FileReader):

    "Reading term information from files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def read_term(self):

        """
        Read a term and its position file offset from the term information file.
        """

        # Read the prefix length and term suffix.

        common = self.read_unsigned_byte()
        suffix = self.read_string()

        self.last_term = self.last_term[:common] + suffix

        # Read the offset delta.

        self.last_offset += self.read_number()

        return self.last_term, self.last_offset

    def go_to_term(self, term, offset, info_offset):

        "Seek past the entry for 'term' having 'offset' to 'info_offset'."

        self.f.seek(info_offset)
        self.last_term = term
        self.last_offset = offset

class TermIndexWriter(TermWriter):

    "Writing term dictionary index details to files."

    def reset(self):
        TermWriter.reset(self)
        self.last_info_offset = 0

    def write_term(self, term, offset, info_offset):

        """
        Write the given 'term' and its position file 'offset' to the term
        dictionary index file, along with the 'info_offset' in the term
        information file.
        """

        TermWriter.write_term(self, term, offset)

        # Write the information file offset delta.

        self.write_number(info_offset - self.last_info_offset)
        self.last_info_offset = info_offset

class TermIndexReader(TermReader):

    "Reading term dictionary index details from files."

    def reset(self):
        TermReader.reset(self)
        self.last_info_offset = 0

    def read_term(self):

        """
        Read a term, its position file offset, and its term information file
        offset from the term dictionary index file.
        """

        term, offset = TermReader.read_term(self)

        # Read the offset delta.

        self.last_info_offset += self.read_number()

        return term, offset, self.last_info_offset

class TermDictionaryWriter:

    "Writing term dictionaries."

    def __init__(self, info_writer, index_writer, interval):
        self.info_writer = info_writer
        self.index_writer = index_writer
        self.interval = interval
        self.entry = 0

    def write_term(self, term, offset):

        """
        Write the given 'term' and its position file 'offset' to the term
        information file and optionally to the index, making a dictionary entry.
        """

        info_offset = self.info_writer.write_term(term, offset)

        if self.entry % self.interval == 0:
            self.index_writer.write_term(term, offset, info_offset)

        self.entry += 1

    def close(self):
        self.info_writer.close()
        self.index_writer.close()

class TermDictionaryReader:

    "Reading term dictionaries."

    def __init__(self, info_reader, index_reader):
        self.info_reader = info_reader
        self.index_reader = index_reader

        self.terms = []
        try:
            while 1:
                self.terms.append(self.index_reader.read_term())
        except EOFError:
            pass

        # Large numbers for ordering purposes.

        self.max_offset = self.terms[-1][1]
        self.max_info_offset = self.terms[-1][2]

    def find(self, term):

        "Find the position file offset of 'term' from the term dictionary."

        i = bisect_right(self.terms, (term, self.max_offset, self.max_info_offset)) - 1

        # Get the entry position providing the term or one preceding it.

        if i == -1:
            return None

        found_term, offset, info_offset = self.terms[i]

        # Where the term is found immediately, return the offset.

        if term == found_term:
            return offset

        # Otherwise, seek past the index term's entry in the information file
        # and scan for the desired term.

        else:
            self.info_reader.go_to_term(found_term, offset, info_offset)
            try:
                while term > found_term:
                    found_term, offset = self.info_reader.read_term()
            except EOFError:
                pass

            # If the term is found, return the offset.

            if term == found_term:
                return offset
            else:
                return None

    def close(self):
        self.info_reader.close()
        self.index_reader.close()

# vim: tabstop=4 expandtab shiftwidth=4
